# -*- coding: utf-8 -*-

import os
import json
import time
import logging
from io import BytesIO

from tornado import web
from tornado import gen

from litedfs.data.handlers.base import BaseHandler, BaseSocketHandler
from litedfs.data.utils.registrant import Registrant
from litedfs.data.utils.common import file_sha1sum, file_md5sum, Errors, splitall
from litedfs.data.config import CONFIG

LOG = logging.getLogger("__name__")


def parse_node_ids(ids_str):
    node_ids = []
    node_ids_tmp = ids_str.split(",")
    for i in node_ids_tmp:
        if i:
            node_ids.append(i)
    return node_ids


class CreateBlockHandler(BaseHandler):
    @gen.coroutine
    def post(self):
        result = {"result": Errors.OK}
        try:
            file_name = self.get_argument("name", "")
            block_id = self.get_argument("block", "")
            node_ids = parse_node_ids(self.get_argument("ids", ""))
            file_body = self.request.files['up_file'][0]["body"]
            if file_name and block_id:
                dir_path = os.path.join(CONFIG["data_path"], "files", file_name[:2], file_name[2:4])
                if not os.path.exists(dir_path):
                    os.makedirs(dir_path)
                file_path = os.path.join(dir_path, "%s_%s.blk" % (file_name, block_id))
                fp = open(file_path, "wb")
                fp.write(file_body)
                fp.close()
                Registrant.instance().replicate_block(file_name, BytesIO(file_body), block_id, node_ids)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class DownloadBlockHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            file_name = self.get_argument("name", "")
            block_id = self.get_argument("block", "")
            if file_name and block_id:
                file_path = os.path.join(CONFIG["data_path"], "files", file_name[:2], file_name[2:4], "%s_%s.blk" % (file_name, block_id))
                if os.path.exists(file_path) and os.path.isfile(file_path):
                    buf_size = 64 * 1024
                    self.set_header('Content-Type', 'application/octet-stream')
                    self.set_header('Content-Disposition', 'attachment; filename=%s_%s.blk' % (file_name, block_id))
                    with open(file_path, 'rb') as f:
                        while True:
                            data = f.read(buf_size)
                            if not data:
                                break
                            self.write(data)
                            self.flush()
                            yield gen.moment
                    self.finish()
                    return
                else:
                    Errors.set_result_error("BlockNotExists", result)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()



# -*- coding: utf-8 -*-

import os
import json
import time
import random
import urllib
import logging
from uuid import uuid4

from tornado import web
from tornado import gen

from litedfs.name.handlers.base import BaseHandler, BaseSocketHandler
from litedfs.name.utils.fs_core import FileSystemTree, InvalidValueError, SameNameExistsError, TargetPathMustDirectoryError, TargetPathNotExistsError, SourcePathNotExistsError, FileNotExistsError
from litedfs.name.utils.listener import Connection
from litedfs.name.utils.common import file_sha1sum, file_md5sum, Errors, splitall
from litedfs.name.config import CONFIG

LOG = logging.getLogger("__name__")


class GenerateFileBlockListHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            file_size = int(self.get_argument("size", "0"))
            replica = int(self.get_argument("replica", "1"))
            block_size = CONFIG["block_size"]
            data_nodes = Connection.get_node_infos()
            for i in data_nodes:
                data_node = data_nodes[i]
                if data_node[0] == "127.0.0.1":
                    host_parts = urllib.parse.urlsplit("//" + self.request.host)
                    data_node[0] = host_parts.hostname
            if len(data_nodes) > 0:
                data_node_ids = list(data_nodes.keys())
                blocks = []
                block_id = 0
                if replica < 1:
                    replica = 1
                if replica > len(data_nodes):
                    replica = len(data_nodes)
                while file_size > block_size:
                    blocks.append((block_id, block_size, random.sample(data_node_ids, replica)))
                    file_size -= block_size
                    block_id += 1
                if file_size > 0:
                    blocks.append((block_id, file_size, random.sample(data_node_ids, replica)))
                result["data_nodes"] = data_nodes
                result["blocks"] = blocks
                result["id"] = str(uuid4())
            else:
                Errors.set_result_error("AllDataNodeOffline", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class CreateFileHandler(BaseHandler):
    @gen.coroutine
    def post(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            file_size = int(self.get_json_argument("size", "0"))
            file_path = self.get_json_argument("path", "")
            file_id = self.get_json_argument("id", "")
            replica = int(self.get_json_argument("replica", "1"))
            blocks = self.get_json_argument("blocks", [])
            if file_path and file_id and replica:
                success = FileSystemTree.instance().create(file_path, {"size": file_size, "id": file_id, "replica": replica, "blocks": blocks})
                if not success:
                    Errors.set_result_error("OperationFailed", result)
            else:
                Errors.set_result_error("InvalidParameters", result)
        except InvalidValueError as e:
            LOG.error(e)
            Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class MoveFileDirectoryHandler(BaseHandler):
    @gen.coroutine
    def put(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            source_path = self.get_json_argument("source_path", "")
            target_path = self.get_json_argument("target_path", "")
            if source_path and target_path:
                success = FileSystemTree.instance().move(source_path, target_path)
                if not success:
                    Errors.set_result_error("OperationFailed", result)
            else:
                Errors.set_result_error("InvalidParameters", result)
        except InvalidValueError as e:
            LOG.error(e)
            Errors.set_result_error("InvalidParameters", result)
        except SameNameExistsError as e:
            LOG.error(e)
            Errors.set_result_error("SameNameExists", result)
        except TargetPathMustDirectoryError as e:
            LOG.error(e)
            Errors.set_result_error("TargetPathMustDirectory", result)
        except TargetPathNotExistsError as e:
            LOG.error(e)
            Errors.set_result_error("TargetPathNotExists", result)
        except SourcePathNotExistsError as e:
            LOG.error(e)
            Errors.set_result_error("SourcePathNotExists", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class RenameFileDirectoryHandler(BaseHandler):
    @gen.coroutine
    def put(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            file_path = self.get_json_argument("path", "")
            new_name = self.get_json_argument("new_name", "")
            if file_path and new_name:
                success = FileSystemTree.instance().rename(file_path, new_name)
                if not success:
                    Errors.set_result_error("OperationFailed", result)
            else:
                Errors.set_result_error("InvalidParameters", result)
        except InvalidValueError as e:
            LOG.error(e)
            Errors.set_result_error("InvalidParameters", result)
        except SameNameExistsError as e:
            LOG.error(e)
            Errors.set_result_error("SameNameExists", result)
        except FileNotExistsError as e:
            LOG.error(e)
            Errors.set_result_error("FileNotExists", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class DeleteFileHandler(BaseHandler):
    @gen.coroutine
    def delete(self):
        result = {"result": Errors.OK}
        try:
            file_path = self.get_argument("path", "")
            if file_path:
                success = FileSystemTree.instance().delete(file_path)
                if not success:
                    Errors.set_result_error("OperationFailed", result)
            else:
                Errors.set_result_error("InvalidParameters", result)
        except InvalidValueError as e:
            LOG.error(e)
            Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class CreateDirectoryHandler(BaseHandler):
    @gen.coroutine
    def post(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            dir_path = self.get_json_argument("path", "")
            if dir_path:
                success = FileSystemTree.instance().makedirs(dir_path)
                if not success:
                    Errors.set_result_error("OperationFailed", result)
            else:
                Errors.set_result_error("InvalidParameters", result)
        except InvalidValueError as e:
            LOG.error(e)
            Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class DeleteDirectoryHandler(BaseHandler):
    @gen.coroutine
    def delete(self):
        result = {"result": Errors.OK}
        try:
            dir_path = self.get_argument("path", "")
            if dir_path:
                success = FileSystemTree.instance().delete(dir_path)
                if not success:
                    Errors.set_result_error("OperationFailed", result)
            else:
                Errors.set_result_error("InvalidParameters", result)
        except InvalidValueError as e:
            LOG.error(e)
            Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class ListDirectoryHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            dir_path = self.get_argument("path", "")
            if dir_path:
                result["children"] = FileSystemTree.instance().list_dir(dir_path)
            else:
                Errors.set_result_error("InvalidParameters", result)
        except InvalidValueError as e:
            LOG.error(e)
            Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class GetFileBlockInfoHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            file_path = self.get_argument("path", "")
            data_nodes = Connection.get_node_infos()
            for i in data_nodes:
                data_node = data_nodes[i]
                if data_node[0] == "127.0.0.1":
                    host_parts = urllib.parse.urlsplit("//" + self.request.host)
                    data_node[0] = host_parts.hostname
            if file_path:
                file_info = FileSystemTree.instance().get_file_info(file_path)
                if file_info:
                    result["file_info"] = file_info
                    result["data_nodes"] = data_nodes
                else:
                    Errors.set_result_error("FileNotExists", result)
            else:
                Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()



# -*- coding: utf-8 -*-

import os
import sys
import json
import logging

cwd = os.path.split(os.path.realpath(__file__))[0]
sys.path.insert(0, os.path.split(cwd)[0])

from litedfs.name.utils.fs_core import FileSystemTree
from litedfs.name.config import CONFIG
from litedfs.name import logger

LOG = logging.getLogger(__name__)


if __name__ == "__main__":
    logger.config_logging(file_name = "test_file_system_tree.log",
                          log_level = "DEBUG",
                          dir_name = os.path.join(cwd, "logs"),
                          day_rotate = False,
                          when = "D",
                          interval = 1,
                          max_size = 20,
                          backup_count = 5,
                          console = True)

    LOG.debug("test start")
    
    try:
        CONFIG["data_path"] = "."

        fs = FileSystemTree()
        # fs.create("/a/b/c/d/e/f/g.txt", {"size": 100})
        # fs.create("/a/b/c/d/e/f.txt", {"size": 100})
        # fs.makedirs("/a/b/c/d/e/f/g/h")
        # fs.move("/a/b/c/d/e/f.txt", "/a/b")
        # fs.delete("/a/b/f.txt")
        LOG.debug("%s", json.dumps(fs.cache, indent = 4))
        LOG.debug("%s", json.dumps(fs.get_info("/"), indent = 4))
        # LOG.debug("%s", json.dumps(fs.get_info("/a/b/c/d"), indent = 4))
        # LOG.debug("%s", json.dumps(fs.list_dir("/a/b/c/d/e", recursive = True), indent = 4))
    except Exception as e:
        LOG.exception(e)

    LOG.debug("test end")



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-

import os
import re
import sys
import json
import time
import argparse
import logging
import random
from io import BytesIO

import requests

from litedfs.version import __version__

parser = argparse.ArgumentParser(prog = 'litedfs')

# common arguments
parser.add_argument("address", help = "name node address, host:port")
parser.add_argument("-w", "--column_width", help = "column max width", type = int, default = 0)
parser.add_argument("-v", "--version", action = 'version', version = '%(prog)s ' + __version__)
subparsers = parser.add_subparsers(dest = "object", help = 'sub-command help')

# operate with file
parser_file = subparsers.add_parser("file", help = "operate with file API")
subparsers_file = parser_file.add_subparsers(dest = "operation", help = 'sub-command file help')

parser_file_create = subparsers_file.add_parser("create", help = "create file")
parser_file_create.add_argument("-l", "--local-path", required = True, help = "local file path", default = "")
parser_file_create.add_argument("-r", "--remote-path", required = True, help = "remote file path", default = "")
parser_file_create.add_argument("-R", "--replica", help = "replica count", type = int, default = 1)

parser_file_delete = subparsers_file.add_parser("delete", help = "delete file")
parser_file_delete.add_argument("-r", "--remote-path", required = True, help = "remote file path", default = "")

parser_file_move = subparsers_file.add_parser("move", help = "move file")
parser_file_move.add_argument("-s", "--source-path", required = True, help = "source file path", default = "")
parser_file_move.add_argument("-t", "--target-path", required = True, help = "target directory path", default = "")

parser_file_rename = subparsers_file.add_parser("rename", help = "rename file")
parser_file_rename.add_argument("-r", "--remote-path", required = True, help = "remote file path", default = "")
parser_file_rename.add_argument("-n", "--new-name", required = True, help = "new file name", default = "")

parser_file_download = subparsers_file.add_parser("download", help = "download file")
parser_file_download.add_argument("-l", "--local-path", required = True, help = "local file path", default = "")
parser_file_download.add_argument("-r", "--remote-path", required = True, help = "remote file path", default = "")

parser_file_info = subparsers_file.add_parser("info", help = "get file's info")
parser_file_info.add_argument("-r", "--remote-path", required = True, help = "remote file path", default = "")

# operate with directory
parser_directory = subparsers.add_parser("directory", help = "operate with directory API")
subparsers_directory = parser_directory.add_subparsers(dest = "operation", help = 'sub-command file help')

parser_directory_create = subparsers_directory.add_parser("create", help = "create directory")
parser_directory_create.add_argument("-r", "--remote-path", required = True, help = "remote directory path", default = "")

parser_directory_delete = subparsers_directory.add_parser("delete", help = "delete directory")
parser_directory_delete.add_argument("-r", "--remote-path", required = True, help = "remote directory path", default = "")

parser_directory_move = subparsers_directory.add_parser("move", help = "move directory")
parser_directory_move.add_argument("-s", "--source-path", required = True, help = "source directory path", default = "")
parser_directory_move.add_argument("-t", "--target-path", required = True, help = "target directory path", default = "")

parser_directory_rename = subparsers_directory.add_parser("rename", help = "rename directory")
parser_directory_rename.add_argument("-r", "--remote-path", required = True, help = "remote directory path", default = "")
parser_directory_rename.add_argument("-n", "--new-name", required = True, help = "new directory name", default = "")

parser_directory_list = subparsers_directory.add_parser("list", help = "list directory's children")
parser_directory_list.add_argument("-r", "--remote-path", required = True, help = "remote directory path", default = "")

args = parser.parse_args()


def print_table_result(data, fields):
    fields.insert(0, "#")
    field_length_map = {}
    lines = []
    num = 1
    column_max_width = args.column_width
    for field in fields:
        field_length_map[field] = len(field)
    for item in data:
        line = []
        for field in fields:
            if field == "#":
                line.append(str(num))
            else:
                v = str(item[field]) if field in item else ""
                v_len = len(v)
                if column_max_width > 0 and v_len > column_max_width:
                    v_len = column_max_width
                    v = v[:v_len]
                if v_len > field_length_map[field]:
                    field_length_map[field] = v_len
                line.append(v)
        lines.append(tuple(line))
        num += 1
    field_length_map["#"] = len(str(num))
    format_str = ""
    for field in fields:
        field_len = field_length_map[field]
        if field == "#":
            format_str += "%" + " %s" % field_len + "s | "
        else:
            format_str += "%" + "-%s" % field_len + "s | "
    format_str = format_str[:-3]
    print(format_str % tuple(fields))
    for line in lines:
        print(format_str % line)


def main():
    try:
        address = args.address
        object = args.object
        operation = args.operation
        url = "http://%s/%s/%s" % (address, object, operation)
        if address:
            if object == "file":
                if operation == "create":
                    if os.path.exists(args.local_path) and os.path.isfile(args.local_path):
                        success = True
                        file_size = os.stat(args.local_path).st_size
                        block_list_url = "http://%s/file/block/list?size=%s&replica=%s" % (address, file_size, args.replica)
                        r = requests.get(block_list_url)
                        if r.status_code == 200:
                            data = r.json()
                            if "result" in data and data["result"] == "ok":
                                data_nodes = data["data_nodes"]
                                fp = open(args.local_path, "rb")
                                for block in data["blocks"]:
                                    block_id = block[0]
                                    block_size = block[1]
                                    block_content = BytesIO()
                                    block_content.write(fp.read(block_size))
                                    node_id = block[2][0]
                                    node_ids = ",".join([str(b) for b in block[2][1:]])
                                    data_node = data_nodes[str(node_id)]
                                    block_create_url = "http://%s:%s/block/create" % (data_node[0], data_node[1])
                                    block_content.seek(0)
                                    files = {'up_file': ("up_file", block_content, b"text/plain")}
                                    values = {"name": data["id"], "block": block_id, "ids": node_ids}
                                    r = requests.post(block_create_url, files = files, data = values)
                                    if r.status_code == 200:
                                        d = r.json()
                                        if "result" in d and d["result"] != "ok":
                                            print("error:\ncode: %s\ncontent: %s" % (r.status_code, r.content))
                                            success = False
                                            break
                                    if not success:
                                        break
                                fp.close()
                                if success:
                                    json_data = {
                                        "size": file_size,
                                        "path": args.remote_path,
                                        "id": data["id"],
                                        "replica": args.replica,
                                        "blocks": data["blocks"],
                                    }
                                    r = requests.post(url, json = json_data)
                                    if r.status_code == 200:
                                        d = r.json()
                                        if "result" in d and d["result"] == "ok":
                                            print("create file[%s] success" % args.remote_path)
                                        else:
                                            print("create file[%s] failed: %s" % (args.remote_path, d["result"]))
                                    else:
                                        print("error:\ncode: %s\ncontent: %s" % (r.status_code, r.content))
                                else:
                                    print("create file[%s] failed" % args.remote_path)
                        else:
                            print("error:\ncode: %s\ncontent: %s" % (r.status_code, r.content))
                    else:
                        print("file[%s] not exists" % args.local_path)
                elif operation == "delete":
                    if args.remote_path:
                        url += "?path=%s" % args.remote_path
                        r = requests.delete(url)
                        if r.status_code == 200:
                            data = r.json()
                            if "result" in data and data["result"] == "ok":
                                print("delete file[%s] success" % args.remote_path)
                            else:
                                print("delete file[%s] failed: %s" % (args.remote_path, data["result"]))
                        else:
                            print("error:\ncode: %s\ncontent: %s" % (r.status_code, r.content))
                elif operation == "move":
                    if args.source_path and args.target_path:
                        json_data = {"source_path": args.source_path, "target_path": args.target_path}
                        r = requests.put(url, json = json_data)
                        if r.status_code == 200:
                            data = r.json()
                            if "result" in data and data["result"] == "ok":
                                print("move file[%s] to %s success" % (args.source_path, args.target_path))
                            else:
                                print("move file[%s] to %s failed: %s" % (args.source_path, args.target_path, data["result"]))
                        else:
                            print("error:\ncode: %s\ncontent: %s" % (r.status_code, r.content))
                elif operation == "rename":
                    if args.remote_path and args.new_name:
                        json_data = {"path": args.remote_path, "new_name": args.new_name}
                        r = requests.put(url, json = json_data)
                        if r.status_code == 200:
                            data = r.json()
                            if "result" in data and data["result"] == "ok":
                                print("rename file[%s] to %s success" % (args.remote_path, args.new_name))
                            else:
                                print("rename file[%s] to %s failed: %s" % (args.remote_path, args.new_name, data["result"]))
                        else:
                            print("error:\ncode: %s\ncontent: %s" % (r.status_code, r.content))
                elif operation == "download":
                    if not os.path.exists(args.local_path):
                        success = True
                        block_info_url = "http://%s/file/block/info?path=%s" % (address, args.remote_path)
                        r = requests.get(block_info_url)
                        if r.status_code == 200:
                            data = r.json()
                            if "result" in data and data["result"] == "ok":
                                data_nodes = data["data_nodes"]
                                file_info = data["file_info"]
                                fp = open(args.local_path, "wb")
                                for block in file_info["blocks"]:
                                    block_id = block[0]
                                    block_size = block[1]
                                    node_id = random.choice(block[2])
                                    data_node = data_nodes[str(node_id)]
                                    block_download_url = "http://%s:%s/block/download?name=%s&block=%s" % (data_node[0], data_node[1], file_info["id"], block_id)
                                    r = requests.get(block_download_url)
                                    if r.status_code == 200:
                                        fp.write(r.content)
                                    else:
                                        print("error:\ncode: %s\ncontent: %s" % (r.status_code, r.content))
                                        success = False
                                        break
                                if success:
                                    fp.close()
                                    print("download file[%s => %s] success" % (args.remote_path, args.local_path))
                                else:
                                    print("download file[%s => %s] failed" % (args.remote_path, args.local_path))
                            else:
                                print("download file[%s] failed: %s" % (args.remote_path, data["result"]))
                        else:
                            print("error:\ncode: %s\ncontent: %s" % (r.status_code, r.content))
                    else:
                        print("local file[%s] already exists" % args.local_path)
                elif operation == "info":
                    if args.remote_path:
                        block_info_url = "http://%s/file/block/info?path=%s" % (address, args.remote_path)
                        r = requests.get(block_info_url)
                        if r.status_code == 200:
                            data = r.json()
                            if "result" in data and data["result"] == "ok":
                                print(json.dumps(data, indent = 4, sort_keys = True))
                            else:
                                print("get file[%s]'s info failed: %s" % (args.remote_path, data["result"]))
                        else:
                            print("error:\ncode: %s\ncontent: %s" % (r.status_code, r.content))
            elif object == "directory":
                if operation == "create":
                    if args.remote_path:
                        json_data = {"path": args.remote_path}
                        r = requests.post(url, json = json_data)
                        if r.status_code == 200:
                            data = r.json()
                            if "result" in data and data["result"] == "ok":
                                print("create directory[%s] success" % args.remote_path)
                            else:
                                print("create directory[%s] failed: %s" % (args.remote_path, data["result"]))
                        else:
                            print("error:\ncode: %s\ncontent: %s" % (r.status_code, r.content))
                elif operation == "delete":
                    if args.remote_path:
                        url += "?path=%s" % args.remote_path
                        r = requests.delete(url)
                        if r.status_code == 200:
                            data = r.json()
                            if "result" in data and data["result"] == "ok":
                                print("delete directory[%s] success" % args.remote_path)
                            else:
                                print("delete directory[%s] failed: %s" % (args.remote_path, data["result"]))
                        else:
                            print("error:\ncode: %s\ncontent: %s" % (r.status_code, r.content))
                elif operation == "move":
                    if args.source_path and args.target_path:
                        json_data = {"source_path": args.source_path, "target_path": args.target_path}
                        r = requests.put(url, json = json_data)
                        if r.status_code == 200:
                            data = r.json()
                            if "result" in data and data["result"] == "ok":
                                print("move directory[%s] to %s success" % (args.source_path, args.target_path))
                            else:
                                print("move directory[%s] to %s failed: %s" % (args.source_path, args.target_path, data["result"]))
                        else:
                            print("error:\ncode: %s\ncontent: %s" % (r.status_code, r.content))
                elif operation == "rename":
                    if args.remote_path and args.new_name:
                        json_data = {"path": args.remote_path, "new_name": args.new_name}
                        r = requests.put(url, json = json_data)
                        if r.status_code == 200:
                            data = r.json()
                            if "result" in data and data["result"] == "ok":
                                print("rename directory[%s] to %s success" % (args.remote_path, args.new_name))
                            else:
                                print("rename directory[%s] to %s failed: %s" % (args.remote_path, args.new_name, data["result"]))
                        else:
                            print("error:\ncode: %s\ncontent: %s" % (r.status_code, r.content))
                elif operation == "list":
                    if args.remote_path:
                        url += "?path=%s" % args.remote_path
                        r = requests.get(url)
                        if r.status_code == 200:
                            data = r.json()
                            if "result" in data and data["result"] == "ok":
                                print_table_result(
                                    data["children"],
                                    ["id", "type", "size", "name"]
                                )
                            else:
                                print("list directory[%s] failed: %s" % (args.remote_path, data["result"]))
                        else:
                            print("error:\ncode: %s\ncontent: %s" % (r.status_code, r.content))

                        
    except Exception as e:
        logging.error(logging.traceback.format_exc())


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

import os
import json
import logging

from litedfs.name.utils.append_log import AppendLogJson
from litedfs.name.utils.listener import Connection
from litedfs.name.utils.common import file_sha1sum, file_md5sum, Errors, splitall
from litedfs.name.config import CONFIG

LOG = logging.getLogger(__name__)


class F(object):
    children = "c"
    type = "t"
    file = "f"
    dir = "d"
    info = "i"
    cmd = "c"
    new_name = "n"
    path = "p"
    source_path = "s"
    target_path = "t"


class C(object):
    create = "c"
    makedirs = "md"
    delete = "d"
    rename = "r"
    move = "m"
    copy = "cp"


class InvalidValueError(Exception):
    def __init__(self, message):
        self.message = message


class SameNameExistsError(Exception):
    def __init__(self, message):
        self.message = message


class FileNotExistsError(Exception):
    def __init__(self, message):
        self.message = message


class TargetPathMustDirectoryError(Exception):
    def __init__(self, message):
        self.message = message


class SourcePathNotExistsError(Exception):
    def __init__(self, message):
        self.message = message


class TargetPathNotExistsError(Exception):
    def __init__(self, message):
        self.message = message


class FileSystemTree(object):
    _instance = None
    name = "file_system_tree"

    def __new__(cls):
        if not cls._instance:
            cls._instance = object.__new__(cls)
            cls._instance.cache = {F.children: {}, F.type: "root"}
            cls._instance.editlog = None
            cls._instance.load_fsimage()
            cls._instance.load_editlog()
            cls._instance.dump_fsimage()
            cls._instance.editlog = AppendLogJson(os.path.join(CONFIG["data_path"], "editlog"))
        return cls._instance

    @classmethod
    def instance(cls):
        return cls._instance

    def create(self, file_path, file_info):
        result = False
        dir_path, file_name = os.path.split(file_path)
        parent = self.makedirs(dir_path)
        if parent:
            parent[F.children][file_name] = {F.type: F.file, F.info: file_info}
            if self.editlog:
                self.editlog.writeline({F.cmd: C.create, F.path: file_path, F.info: file_info})
            result = True
        return result

    def delete(self, file_path):
        result = True
        _, name = os.path.split(file_path)
        exists, file_type, file, parent = self.get_info(file_path)
        if exists:
            del parent[F.children][name]
            if file[F.type] == F.file:
                file_id = file[F.info]["id"]
                task = {"command": "delete", "name": file_id}
                for i in Connection.id_decompress:
                    if i not in Connection.tasks:
                        Connection.tasks[i] = [task]
                    else:
                        Connection.tasks[i].append(task)
            elif file[F.type] == F.dir:
                self.delete_files(file)
            if self.editlog:
                self.editlog.writeline({F.cmd: C.delete, F.path: file_path})
        return result

    def delete_files(self, file):
        if F.children in file and file[F.children]:
            for name in file[F.children]:
                child = file[F.children][name]
                self.delete_files(child)
        else:
            if file[F.type] == F.file:
                file_id = file[F.info]["id"]
                task = {"command": "delete", "name": file_id}
                for i in Connection.id_decompress:
                    if i not in Connection.tasks:
                        Connection.tasks[i] = [task]
                    else:
                        Connection.tasks[i].append(task)
                LOG.debug("delete file: %s", file)

    def get_file_info(self, file_path):
        result = False
        exists, file_type, file, _ = self.get_info(file_path)
        if exists:
            result = file[F.info]
        return result

    def rename(self, file_path, new_name):
        result = True
        if self.is_valid_name(new_name):
            _, name = os.path.split(file_path)
            exists, file_type, file, parent = self.get_info(file_path)
            if exists:
                if new_name not in parent[F.children]:
                    parent[F.children][new_name] = file
                    del parent[F.children][name]
                    if self.editlog:
                        self.editlog.writeline({F.cmd: C.rename, F.path: file_path, F.new_name: new_name})
                else:
                    raise SameNameExistsError("same file name exists: %s" % new_name)
            else:
                raise FileNotExistsError("file not exists: %s" % file_path)
        else:
            raise InvalidValueError("invailed file name: %s" % new_name)
        return result

    def is_valid_name(self, name):
        result = True
        if "/" in name:
            result = False
        return result

    def move(self, source_path, target_path):
        result = False
        _, name = os.path.split(source_path)
        source_exists, _, source_file, source_parent = self.get_info(source_path)
        if source_exists:
            target_exists, target_type, target_file, _ = self.get_info(target_path)
            if target_exists:
                if target_type in (F.dir, "root"):
                    if name not in target_file[F.children]:
                        target_file[F.children][name] = source_file
                        del source_parent[F.children][name]
                        if self.editlog:
                            self.editlog.writeline({F.cmd: C.move, F.source_path: source_path, F.target_path: target_path})
                        result = True
                    else:
                        raise SameNameExistsError("same file name exists: %s" % name)
                else:
                    raise TargetPathMustDirectoryError("target path must be directory: %s" % target_path)
            else:
                raise TargetPathNotExistsError("target path not exists: %s" % target_path)
        else:
            raise SourcePathNotExistsError("source path not exists: %s" % source_path)
        return result

    def copy(self, source_path, target_path):
        result = False
        _, name = os.path.split(source_path)
        source_exists, _, source_file, source_parent = self.get_info(source_path)
        if source_exists:
            target_exists, target_type, target_file, _ = self.get_info(target_path)
            if target_exists:
                if target_type == F.dir:
                    if name not in target_file[F.children]:
                        target_file[F.children][name] = source_file
                        if self.editlog:
                            self.editlog.writeline({F.cmd: C.copy, F.source_path: source_path, F.target_path: target_path})
                        result = True
                    else:
                        raise SameNameExistsError("same file name exists: %s" % name)
                else:
                    raise TargetPathMustDirectoryError("target path must be directory: %s" % target_path)
            else:
                raise TargetPathNotExistsError("target path not exists: %s" % target_path)
        else:
            raise SourcePathNotExistsError("source path not exists: %s" % source_path)
        return result

    def list_dir(self, directory_path, recursive = False):
        result = []
        exists, file_type, file, _ = self.get_info(directory_path)
        if exists and file_type in (F.dir, "root"):
            if recursive:
                result = file[F.children]
            else:
                names = list(file[F.children].keys())
                names.sort()
                for name in names:
                    file_type = file[F.children][name][F.type]
                    child = {
                        "name": name,
                    }
                    if file_type == F.file:
                        child["type"] = "file"
                        child["size"] = file[F.children][name][F.info]["size"]
                        child["id"] = file[F.children][name][F.info]["id"]
                    elif file_type == F.dir:
                        child["type"] = "directory"
                        child["size"] = 0
                        child["id"] = ""
                    result.append(child)
            LOG.debug("list_dir: %s", result)
        return result

    def exists(self, file_path):
        result, _, _, _ = self.get_info(directory_path)
        return result

    def makedirs(self, directory_path):
        result = False
        path_parts = splitall(directory_path)
        if path_parts[0] != "/":
            raise InvalidValueError("must be absolute path: %s" % directory_path)
        else:
            current_root = self.cache
            for dir_name in path_parts[1:]:
                if dir_name not in current_root[F.children]:
                    current_root[F.children][dir_name] = {F.type: F.dir, F.children: {}}
                current_root = current_root[F.children][dir_name]
            if self.editlog:
                self.editlog.writeline({F.cmd: C.makedirs, F.path: directory_path})
            result = current_root
        return result

    def isdir(self, directory_path):
        result = False
        exists, file_type, _, _ = self.get_info(directory_path)
        if exists and file_type == F.dir:
            result = True
        return result

    def isfile(self, file_path):
        result = False
        exists, file_type, _, _ = self.get_info(file_path)
        if exists and file_type == F.file:
            result = True
        return result

    def get_info(self, file_path):
        result = [True, "", {}, None]
        path_parts = splitall(file_path)
        if path_parts[0] != "/":
            raise InvalidValueError("must be absolute path: %s" % file_path)
        else:
            parent = None
            current_root = self.cache
            for name in path_parts[1:]:
                if name not in current_root[F.children]:
                    result[0] = False
                    break
                else:
                    parent = current_root
                    current_root = current_root[F.children][name]
            if result[0]:
                result[1] = current_root[F.type]
                result[2] = current_root
                result[3] = parent
        return result

    def load_fsimage(self):
        result = False
        try:
            LOG.info("loading fsimage ...")
            fsimage_path = os.path.join(CONFIG["data_path"], "fsimage")
            fsimage = AppendLogJson(fsimage_path)
            for line in fsimage.iterlines():
                if line[F.cmd] == C.create:
                    self.create(line[F.path], line[F.info])
                elif line[F.cmd] == C.makedirs:
                    self.makedirs(line[F.path])
            result = True
        except Exception as e:
            LOG.exception(e)
        return result

    def load_editlog(self):
        result = False
        try:
            LOG.info("loading editlog ...")
            editlog_path = os.path.join(CONFIG["data_path"], "editlog")
            editlog = AppendLogJson(editlog_path)
            for line in editlog.iterlines():
                if line[F.cmd] == C.create:
                    self.create(line[F.path], line[F.info])
                elif line[F.cmd] == C.makedirs:
                    self.makedirs(line[F.path])
                elif line[F.cmd] == C.rename:
                    self.rename(line[F.path], line[F.new_name])
                elif line[F.cmd] == C.delete:
                    self.delete(line[F.path])
                elif line[F.cmd] == C.move:
                    self.move(line[F.source_path], line[F.target_path])
                elif line[F.cmd] == C.copy:
                    self.copy(line[F.source_path], line[F.target_path])
            result = True
        except Exception as e:
            LOG.exception(e)
        return result

    def dump_fsimage(self):
        result = False
        try:
            LOG.info("dumping fsimage ...")
            new_fsimage_path = os.path.join(CONFIG["data_path"], "fsimage.new")
            fsimage_path = os.path.join(CONFIG["data_path"], "fsimage")
            old_fsimage_path = os.path.join(CONFIG["data_path"], "fsimage.old")
            editlog_path = os.path.join(CONFIG["data_path"], "editlog")
            LOG.debug("new fsimage: %s", new_fsimage_path)
            self.new_fsimage = AppendLogJson(new_fsimage_path)
            self.dump_files("/", self.cache)
            if os.path.exists(old_fsimage_path):
                os.remove(old_fsimage_path)
            if os.path.exists(fsimage_path):
                os.rename(fsimage_path, old_fsimage_path)
            if os.path.exists(new_fsimage_path):
                os.rename(new_fsimage_path, fsimage_path)
            if os.path.exists(editlog_path):
                os.remove(editlog_path)
            result = True
        except Exception as e:
            LOG.exception(e)
        return result

    def dump_files(self, file_path, file):
        if F.children in file and file[F.children]:
            for name in file[F.children]:
                child = file[F.children][name]
                self.dump_files(os.path.join(file_path, name), child)
        else:
            if file[F.type] == F.file:
                LOG.debug("find file[%s]: %s", file_path, file)
                self.new_fsimage.writeline({F.cmd: C.create, F.path: file_path, F.info: file[F.info]})
            elif file[F.type] == F.dir:
                LOG.debug("find directory[%s]: %s", file_path, file)
                self.new_fsimage.writeline({F.cmd: C.makedirs, F.path: file_path})

    def close(self):
        pass



# -*- coding: UTF-8 -*-

import os
import re
import sys
import json
import time
import argparse
import logging
import random
from io import BytesIO

import requests

LOG = logging.getLogger(__name__)


class LiteDFSClient(object):
    def __init__(self, host, port):
        self.host = host
        self.port = port
        self.base_url = "http://%s:%s" % (self.host, self.port)

    def create_file(self, local_path, remote_path, replica = 1):
        result = False
        try:
            if os.path.exists(local_path) and os.path.isfile(local_path):
                success = True
                file_size = os.stat(local_path).st_size
                block_list_url = "%s/file/block/list?size=%s&replica=%s" % (self.base_url, file_size, replica)
                r = requests.get(block_list_url)
                if r.status_code == 200:
                    data = r.json()
                    if "result" in data and data["result"] == "ok":
                        data_nodes = data["data_nodes"]
                        fp = open(local_path, "rb")
                        for block in data["blocks"]:
                            block_id = block[0]
                            block_size = block[1]
                            block_content = BytesIO()
                            block_content.write(fp.read(block_size))
                            node_id = block[2][0]
                            node_ids = ",".join([str(b) for b in block[2][1:]])
                            data_node = data_nodes[str(node_id)]
                            block_create_url = "http://%s:%s/block/create" % (data_node[0], data_node[1])
                            block_content.seek(0)
                            files = {'up_file': ("up_file", block_content, b"text/plain")}
                            values = {"name": data["id"], "block": block_id, "ids": node_ids}
                            r = requests.post(block_create_url, files = files, data = values)
                            if r.status_code == 200:
                                d = r.json()
                                if "result" in d and d["result"] != "ok":
                                    LOG.error("create block failed: %s", d)
                                    success = False
                                    break
                            if not success:
                                break
                        fp.close()
                        if success:
                            json_data = {
                                "size": file_size,
                                "path": remote_path,
                                "id": data["id"],
                                "replica": replica,
                                "blocks": data["blocks"],
                            }
                            r = requests.post("%s/file/create" % self.base_url, json = json_data)
                            if r.status_code == 200:
                                d = r.json()
                                if "result" in d and d["result"] == "ok":
                                    result = True
                                else:
                                    LOG.error("create file[%s] failed: %s", remote_path, d["result"])
                            else:
                                LOG.error("error:\ncode: %s\ncontent: %s", r.status_code, r.content)
                        else:
                            LOG.error("create file[%s] failed", remote_path)
                    else:
                        LOG.error("create file[%s] failed: %s", remote_path, data["result"])
                else:
                    LOG.error("error:\ncode: %s\ncontent: %s", r.status_code, r.content)
            else:
                LOG.error("file[%s] not exists", local_path)
        except Exception as e:
            LOG.exception(e)
        return result

    def delete_file(self, remote_path):
        result = False
        try:
            url = "%s/file/delete?path=%s" % (self.base_url, remote_path)
            r = requests.delete(url)
            if r.status_code == 200:
                data = r.json()
                if "result" in data and data["result"] == "ok":
                    result = True
                else:
                    LOG.error("delete file[%s] failed: %s", remote_path, data["result"])
            else:
                LOG.error("error:\ncode: %s\ncontent: %s", r.status_code, r.content)
        except Exception as e:
            LOG.exception(e)
        return result

    def move_file(self, source_path, target_path):
        result = False
        try:
            url = "%s/file/move" % self.base_url
            json_data = {"source_path": source_path, "target_path": target_path}
            r = requests.put(url, json = json_data)
            if r.status_code == 200:
                data = r.json()
                if "result" in data and data["result"] == "ok":
                    result = True
                else:
                    LOG.error("move file[%s] to %s failed: %s", source_path, target_path, data["result"])
            else:
                LOG.error("error:\ncode: %s\ncontent: %s", r.status_code, r.content)
        except Exception as e:
            LOG.exception(e)
        return result

    def rename_file(self, remote_path, new_name):
        result = False
        try:
            url = "%s/file/rename" % self.base_url
            json_data = {"path": remote_path, "new_name": new_name}
            r = requests.put(url, json = json_data)
            if r.status_code == 200:
                data = r.json()
                if "result" in data and data["result"] == "ok":
                    result = True
                else:
                    LOG.error("rename file[%s] to %s failed: %s", remote_path, new_name, data["result"])
            else:
                LOG.error("error:\ncode: %s\ncontent: %s", r.status_code, r.content)
        except Exception as e:
            LOG.exception(e)
        return result

    def download_file(self, remote_path, local_path):
        result = False
        try:
            if not os.path.exists(local_path):
                success = True
                block_info_url = "%s/file/block/info?path=%s" % (self.base_url, remote_path)
                r = requests.get(block_info_url)
                if r.status_code == 200:
                    data = r.json()
                    if "result" in data and data["result"] == "ok":
                        data_nodes = data["data_nodes"]
                        file_info = data["file_info"]
                        fp = open(local_path, "wb")
                        for block in file_info["blocks"]:
                            block_id = block[0]
                            block_size = block[1]
                            node_id = random.choice(block[2])
                            data_node = data_nodes[str(node_id)]
                            block_download_url = "http://%s:%s/block/download?name=%s&block=%s" % (data_node[0], data_node[1], file_info["id"], block_id)
                            r = requests.get(block_download_url)
                            if r.status_code == 200:
                                fp.write(r.content)
                            else:
                                LOG.error("error:\ncode: %s\ncontent: %s", r.status_code, r.content)
                                success = False
                                break
                        if success:
                            fp.close()
                            result = True
                        else:
                            LOG.error("download file[%s => %s] failed", remote_path, local_path)
                    else:
                        LOG.error("download file[%s] failed: %s", remote_path, data["result"])
                else:
                    LOG.error("error:\ncode: %s\ncontent: %s", r.status_code, r.content)
            else:
                LOG.error("local file[%s] already exists", local_path)
        except Exception as e:
            LOG.exception(e)
        return result

    def info_file(self, remote_path):
        result = False
        try:
            block_info_url = "%s/file/block/info?path=%s" % (self.base_url, remote_path)
            r = requests.get(block_info_url)
            if r.status_code == 200:
                data = r.json()
                if "result" in data and data["result"] == "ok":
                    result = data
                else:
                    LOG.error("get file[%s]'s info failed: %s", remote_path, data["result"])
            else:
                LOG.error("error:\ncode: %s\ncontent: %s", r.status_code, r.content)
        except Exception as e:
            LOG.exception(e)
        return result

    def create_directory(self, remote_path):
        result = False
        try:
            url = "%s/directory/create" % self.base_url
            json_data = {"path": remote_path}
            r = requests.post(url, json = json_data)
            if r.status_code == 200:
                data = r.json()
                if "result" in data and data["result"] == "ok":
                    result = True
                else:
                    LOG.error("create directory[%s] failed: %s", remote_path, data["result"])
            else:
                LOG.error("error:\ncode: %s\ncontent: %s", r.status_code, r.content)
        except Exception as e:
            LOG.exception(e)
        return result

    def delete_directory(self, remote_path):
        result = False
        try:
            url = "%s/directory/delete?path=%s" % (self.base_url, remote_path)
            r = requests.delete(url)
            if r.status_code == 200:
                data = r.json()
                if "result" in data and data["result"] == "ok":
                    result = True
                else:
                    LOG.error("delete directory[%s] failed: %s", remote_path, data["result"])
            else:
                LOG.error("error:\ncode: %s\ncontent: %s", r.status_code, r.content)
        except Exception as e:
            LOG.exception(e)
        return result

    def move_directory(self, source_path, target_path):
        result = False
        try:
            url = "%s/directory/move" % self.base_url
            json_data = {"source_path": source_path, "target_path": target_path}
            r = requests.put(url, json = json_data)
            if r.status_code == 200:
                data = r.json()
                if "result" in data and data["result"] == "ok":
                    result = True
                else:
                    LOG.error("move directory[%s] to %s failed: %s", source_path, target_path, data["result"])
            else:
                LOG.error("error:\ncode: %s\ncontent: %s", r.status_code, r.content)
        except Exception as e:
            LOG.exception(e)
        return result

    def rename_directory(self, remote_path, new_name):
        result = False
        try:
            url = "%s/directory/rename" % self.base_url
            json_data = {"path": remote_path, "new_name": new_name}
            r = requests.put(url, json = json_data)
            if r.status_code == 200:
                data = r.json()
                if "result" in data and data["result"] == "ok":
                    result = True
                else:
                    LOG.error("rename directory[%s] to %s failed: %s", remote_path, new_name, data["result"])
            else:
                LOG.error("error:\ncode: %s\ncontent: %s", r.status_code, r.content)
        except Exception as e:
            LOG.exception(e)
        return result

    def list_directory(self, remote_path):
        result = False
        try:
            url = "%s/directory/list?path=%s" % (self.base_url, remote_path)
            r = requests.get(url)
            if r.status_code == 200:
                data = r.json()
                if "result" in data and data["result"] == "ok":
                    result = data
                else:
                    print("list directory[%s] failed: %s" % (remote_path, data["result"]))
            else:
                print("error:\ncode: %s\ncontent: %s" % (r.status_code, r.content))
        except Exception as e:
            LOG.exception(e)
        return result



# -*- coding: utf-8 -*-

import os
import sys
import json
import logging

cwd = os.path.split(os.path.realpath(__file__))[0]
sys.path.insert(0, os.path.split(cwd)[0])

from litedfs.tool.client import LiteDFSClient
from litedfs.name import logger

LOG = logging.getLogger(__name__)


if __name__ == "__main__":
    logger.config_logging(file_name = "test_client.log",
                          log_level = "DEBUG",
                          dir_name = os.path.join(cwd, "logs"),
                          day_rotate = False,
                          when = "D",
                          interval = 1,
                          max_size = 20,
                          backup_count = 5,
                          console = True)

    LOG.debug("test start")
    
    try:
        c = LiteDFSClient("192.168.199.139", 9000)
        # LOG.debug("create file: %s", c.create_file("./test.log", "/test.log"))
        # LOG.debug("rename file: %s", c.rename_file("/test.log", "test.rename.log"))
        # LOG.debug("move file: %s", c.move_file("/test.rename.log", "/test"))
        # LOG.debug("download file: %s", c.download_file("/test/test.rename.log", "./test.rename.log"))
        # LOG.debug("info file: %s", c.info_file("/test/test.rename.log"))
        # LOG.debug("delete file: %s", c.delete_file("/test/test.rename.log"))

        LOG.debug("create directory: %s", c.create_directory("/new_dir"))
        LOG.debug("rename directory: %s", c.rename_directory("/new_dir", "new_dir_name"))
        LOG.debug("move directory: %s", c.move_directory("/new_dir_name", "/test"))
        LOG.debug("list directory: %s", c.list_directory("/test"))
        LOG.debug("delete directory: %s", c.delete_directory("/test/new_dir_name"))
    except Exception as e:
        LOG.exception(e)

    LOG.debug("test end")



# -*- coding: utf-8 -*-
'''
LiteDFS: distributed file system
'''

from setuptools import setup

from litedfs.version import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name = "litedfs",
    version = __version__,
    description = "LiteDFS: distributed file system",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = "https://github.com/fiefdx/LiteDFS",
    author = "fiefdx",
    author_email = "fiefdx@163.com",
    packages = [
        'litedfs',
        'litedfs.tool',
        'litedfs.name',
        'litedfs.name.db',
        'litedfs.name.handlers',
        'litedfs.name.models',
        'litedfs.name.utils',
        'litedfs.data',
        'litedfs.data.handlers',
        'litedfs.data.utils',
    ],
    entry_points = {
        'console_scripts': [
            'litedfs = litedfs.tool.litedfs:main',
            'litename = litedfs.name.name:main',
            'litedata = litedfs.data.data:main',
        ],
    },
    install_requires = [
        "requests >= 2.22.0",
        "tornado",
        "pyYAML",
        "tinydb",
        "sqlalchemy",
        "tornado_discovery",
    ],
    include_package_data = True,
    license = "MIT",
    classifiers = [
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ]
)



# -*- coding: utf-8 -*-

import json
import logging

from tornado import gen
from tornado.httpclient import AsyncHTTPClient, HTTPRequest
from tornado.tcpclient import TCPClient
from tornado_discovery.registrant import BaseRegistrant
from tornado_discovery.common import Command, Status

from litedfs.data.utils.common import Errors, async_post, delete_file
from litedfs.data.config import CONFIG

LOG = logging.getLogger(__name__)


class Registrant(BaseRegistrant):
    _instance = None

    def __new__(cls, host, port, config, retry_interval = 10, reconnect = True):
        if not cls._instance:
            cls._instance = object.__new__(cls)
            cls._instance.host = host
            cls._instance.port = port
            cls._instance.config = config
            cls._instance.retry_interval = retry_interval
            cls._instance.heartbeat_interval = cls._instance.config.get("heartbeat_interval")
            cls._instance.heartbeat_timeout = cls._instance.config.get("heartbeat_timeout")
            cls._instance.reconnect = reconnect
            cls._instance.tcpclient = TCPClient()
            cls._instance.periodic_heartbeat = None
            cls._instance._stream = None
            cls._instance.heartbeat_data = {}
            cls._instance.registered = False
            cls._instance.data_nodes = {}
            cls._instance.async_client = AsyncHTTPClient()
        return cls._instance

    def __init__(self, host, port, config, retry_interval = 10, reconnect = True):
        pass

    @classmethod
    def instance(cls):
        return cls._instance

    def update_heartbeat_data(self, data = {}):
        self.heartbeat_data.update(data)

    @gen.coroutine
    def replicate_block(self, file_name, block_file, block_id, node_ids):
        result = False
        try:
            if node_ids:
                node_id = node_ids[0]
                if node_id in self.data_nodes:
                    data_node = self.data_nodes[node_id]
                    url = "http://%s:%s/block/create" % (data_node[0], data_node[1])
                    r = yield async_post(self.async_client, url, {"up_file": block_file}, {"name": file_name, "block": block_id, "ids": ",".join(node_ids[1:])})
                    if r.code == 200:
                        data = json.loads(r.body.decode("utf-8"))
                        if "result" in data and data["result"] == Errors.OK:
                            result = True
                            LOG.debug("replicate block to node: %s(%s:%s), result: %s", node_id, data_node[0], data_node[1], result)
                        else:
                            LOG.error("replicate block to node: %s(%s:%s) failed, result: %s", node_id, data_node[0], data_node[1], result)
                    else:
                        LOG.error("replicate block to node: %s(%s:%s) failed, response: %s", node_id, data_node[0], data_node[1], r)
            else:
                result = True
        except Exception as e:
            LOG.exception(e)
        return result

    @gen.coroutine
    def register_service(self):
        try:
            data = {"command": Command.register, "data": self.config.to_dict()}
            self.send_message(data)
            data = yield self.read_message()
            if data["command"] == Command.register:
                if data["data"]["status"] == Status.success:
                    self.registered = True
                    if not self.config.has_key("node_id"):
                        self.config.set("node_id", data["data"]["node_id"])
                        LOG.info("Received new node_id: %s", data["data"]["node_id"])
                    LOG.info("Client Register Received Message: %s", data)
                else:
                    LOG.error("Client Register Failed, Received Message: %s", data)
            else:
                LOG.error("Client Register Failed, Received Wrong Message: %s", data)
        except Exception as e:
            LOG.exception(e)

    @gen.coroutine
    def heartbeat_service(self):
        try:
            message_data = self.config.to_dict()
            message_data.update(self.heartbeat_data)
            data = {"command": Command.heartbeat, "data": message_data}
            self.send_message(data)
            data = yield self.read_message()
            if data["data"]["status"] == Status.success:
                if "data_nodes" in data["data"]:
                    self.data_nodes = data["data"]["data_nodes"]
                if "task" in data["data"]:
                    task = data["data"]["task"]
                    if task["command"] == "delete":
                        yield delete_file(task["name"])
                LOG.debug("Client Received Heartbeat Message: %s", data["data"])
            else:
                LOG.error("Client Received Heartbeat Message: %s", data["data"])
        except Exception as e:
            LOG.exception(e)
